#ifndef G9AF4B594133B40D59D1BF5FB15F04D2D
#define G9AF4B594133B40D59D1BF5FB15F04D2D

#include <cstddef>
#include <optional>
#include <string>
#include <type_traits>

/**
To be able to specialize, a base template is required.
Because it’s never used directly it can be a declaration only.
*/
template<typename>
struct EnumExt;

namespace enum_ext
{

template<typename Derived, typename EnumType>
struct Conversion
{
    static_assert(std::is_enum_v<EnumType>);
    using UnderlyingType = std::underlying_type_t<EnumType>;

    // Everything is static. Instantiating doesn’t make sense, so disable it.
    Conversion() = delete;
    ~Conversion() = delete;

    static UnderlyingType to_underlying_type(EnumType e) noexcept;
    static EnumType from_underlying_type(UnderlyingType i, EnumType fallback) noexcept;
    static bool is_convertible_from(UnderlyingType i) noexcept;

    static EnumType from_string(const std::string& str, EnumType fallback) noexcept;
    static bool is_convertible_from(const std::string& str) noexcept;

    /**
    `Derived` must provide functions with the following signatures
    static std::optional<EnumType> from_underlying_type(UnderlyingType) noexcept;
    static std::string to_string(EnumType);
    static std::optional<EnumType> from_string(const std::string&) noexcept;
    */
};


/** A simple tag that indicates the presence of the ``enumerators`` array. */
struct IterationTag
{
};

}  // namespace enum_ext


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// implementation details below


namespace enum_ext
{

template<typename Derived, typename EnumType>
typename Conversion<Derived, EnumType>::UnderlyingType
        Conversion<Derived, EnumType>::to_underlying_type(EnumType e) noexcept
{
    return static_cast<UnderlyingType>(e);
}

template<typename Derived, typename EnumType>
EnumType Conversion<Derived, EnumType>::from_underlying_type(
        UnderlyingType i, EnumType fallback) noexcept
{
    const auto maybe_enumerator = Derived::from_underlying_type(i);

    if (maybe_enumerator) {
        return *maybe_enumerator;
    }

    return fallback;
}

template<typename Derived, typename EnumType>
bool Conversion<Derived, EnumType>::is_convertible_from(UnderlyingType i) noexcept
{
    return Derived::from_underlying_type(i).has_value();
}

template<typename Derived, typename EnumType>
EnumType Conversion<Derived, EnumType>::from_string(
        const std::string& str, EnumType fallback) noexcept
{
    const auto maybe_enumerator = Derived::from_string(str);

    if (maybe_enumerator) {
        return *maybe_enumerator;
    }

    return fallback;
}

template<typename Derived, typename EnumType>
bool Conversion<Derived, EnumType>::is_convertible_from(const std::string& str) noexcept
{
    return Derived::from_string(str).has_value();
}

}  // namespace enum_ext

#endif
